/*** [ rs232_2_twi.c ] *****************************************************
       27.08.2008      Tab = 3

		 Beschreibung ausgelagert in "rs232_2_twi.txt"

****************************************************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <stdio.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <inttypes.h>
#include <string.h>
#include <ctype.h>

#include "TWI_Master.h"
#include "tsl2550.h"
#include "rs232_2_twi.h"

#define EXTERN
#include "globals.h"

// ---------------------------------------------------------------------------
// wird aufgerufen, wenn ein Byte ber COM eingegangen ist
// hat nur die Aufgabe, den Schlafmodus zu beenden
// das Speichern des eingegangenen Bytes wird von main bernommen
EMPTY_INTERRUPT(USART_RXC_vect);

// ----------------------------------------------------------------------------
int main(void)
{
uint8_t temp;
uint8_t i;
uint8_t chr0_flag;							// am Ende eines Textstrings muss ein CHR0 geschrieben werden

no_echo = 0;									// Echo der Eingabe zum Terminal zurcksenden (TxD-Echo AN)
format = 10;									// Parameter fr itoa(): 2 = binr, 10 = dezimal, 16 = hexadezimal

UBRRL = TEILER;								// RS232 initialisieren
UCSRB |= (1<<RXCIE | 1<<TXEN | 1<<RXEN);// Transmitter und Receiver aktivieren
UCSRC |= (1<<URSEL | 0<<UMSEL) | (1<<UCSZ1) | (1<<UCSZ0);
// URSEL - Achtung bei ATmega 8

MCUCR = 1<<SE;									// Sleep enablen (im Idle-Modus)
ACSR = 1<<ACD;									// Analog comparator disablen
TWI_Master_Initialise();
sei();

help();											// Hilfstext mit Befehlen ausgeben

while (1)
	{
	chr_in = 0;									// Index fr vom Terminal eingelesene Zeichen

// Eingabeschleife ------------------------------------------------------------
// liest die Terminaleingabe ein bis zu einem RET oder zum Ende des Puffers
	do
		{											// warten, bis sich ein Interrupt ereignet
		asm volatile("sleep");				// RXIE oder TWIE beendet das Warten

		temp = 0;								// temp zurcksetzen
		if (UCSRA & (1<<RXC))         	// wenn RXIE den Interrupt ausgelst hat
			{
			temp = UDR;							// ohne den Umweg ber temp wird UDR nicht gelesen ???
			if ((temp > 29) || (temp == 13))// alle Steuerzeichen ausser RETURN ignorieren
				{
				rs232_buf[chr_in++] = temp;// ASCII-Wert einlesen, im Puffer ablegen, Index incr
				}
			}
		}											// einlesen, bis RETURN oder Ende des Zeichenpuffers erreicht
	while((!(temp == 13)) && (chr_in < MAX_BUFFER));

// Auswertung der Eingabe -----------------------------------------------------
// wandelt Ziffernfolgen in rs232_buf in Bytes um,  wirft die Trennzeichen "," raus und schreibt das
// Ergebnis in write_buf
// Kommandos (die mit "-" beginnen) werden separat behandelt.

	chr_out 	 = 0;								// Index fr aus dem Puffer ausgelesene Zeichen.
	chr_twi 	 = 0;								// Index fr auf den TWI-Bus auszugebende/lesende Zeichen
  	chr0_flag = 0;

	while((!(rs232_buf[chr_out] == 13)) && (chr_out <= chr_in))
		{
		if ((chr_out == 0) && (rs232_buf[0] == '-'))	// wenn das erste Zeichen = '-', dann folgen Kommandos
			{
			rs232_buf[1] = tolower(rs232_buf[1]);		// Parameter klein schreiben, dannach nur auf Kleinschreibung prfen
			if (rs232_buf[1] == 'l')						// Parameter prfen
				{
				twi_test();
				break;
				}

			else if (rs232_buf[1] == 'x')				// TWI-Takt auf 100KHz
				{
  				TWBR = (F_CPU / (2 * 100000) - 8);
				show_twi_takt(F_CPU/(16 + 2 * 10) / 1000);
				}

			else if (rs232_buf[1] == '0')	 			// TWBR einstellen
				{
  				TWBR = 10;									// TWBR should be 10 or higher
				show_twi_takt(F_CPU/(16 + 2 * 10) / 1000);
				}

			else if (rs232_buf[1] == '1')
				{
  				TWBR = 20;
				show_twi_takt(F_CPU/(16 + 2 * 20) / 1000);
				}

			else if (rs232_buf[1] == '2')
				{
				TWBR = (40);
				show_twi_takt(F_CPU/(16 + 2 * 40) / 1000);
				}

			else if (rs232_buf[1] == '3')
				{
				TWBR = 64;
				show_twi_takt(F_CPU/(16 + 2 * 64) / 1000);
				}

			else if (rs232_buf[1] == '4')
				{
				TWBR = 128;
				show_twi_takt(F_CPU/(16 + 2 * 128) / 1000);
				}

			else if (rs232_buf[1] == '5')
				{
				TWBR = 255;
				show_twi_takt(F_CPU/(16 + 2 * 255) / 1000);
				}

			else if (rs232_buf[1] == 'b')	// Ausgabe binaer
				{
				format = 2;
				show_format();
				}

			else if (rs232_buf[1] == 'd')	// Ausgabe dezimal
				{
				format = 10;
				show_format();
				}

			else if (rs232_buf[1] == 'h')	// Ausgabe hexadezimal
				{
				format = 16;
				show_format();
				}

			else if (rs232_buf[1] == '?')	// Hilfe ausgeben
				{
				help();
				}

			else if (rs232_buf[1] == 'e')	// Eingabemodus / Echo togglen
				{
				no_echo ^= 0x01;
				show_echo();					// aktuelle Einstellung ausgeben
				}

#ifdef TSL2550
			else if (rs232_buf[1] == 't')
				{
				read_lux();
				}
#endif
			else									// alle anderen Parameter ignorieren
				{
				msg[0] = rs232_buf[1];
				msg[1] = 0;
				puts_p(PSTR("unbekannt: -"));// Fehlermeldung ausgeben
				put_string(msg);
				put_cr();
				}

			chr_out = 0;						// der Auswertung mitteilen, dass keine Ausgabe vorliegt
			break;
			}

		else if (rs232_buf[chr_out] == '+')	// ein '+' behlt den letzten Eintrag im Sendepuffer bei,
			{										// so muss nicht immer Adresse/Kommando neu eingegeben werden.
			chr_out++;							// Index Lesepuffer
			chr_twi++;							// Zhler fr die Anzahl der auszugebenden Zeichen
			}

		else if (isdigit(rs232_buf[chr_out])) // wenn eine Ziffer vorliegt, dann in ein Byte umwandeln lassen
			{
			temp = Zahl_read();				// die aufgerufene Routine kann chr_twi noch ndern, daher erst in Variable
			write_buf[chr_twi] = temp;		// einlesen und dann dem Array zuweisen !!
			chr_twi++;
			}

		else if(rs232_buf[chr_out] == ',')// ein Komma einfach bergehen ...
			{
			chr_out++;							// Index Lesepuffer incrementieren, Schreibpuffer nicht ndern
			}

		else										// wenn das nchste Zeichen keine Ziffer ist,
			{										// dann wird der Rest als Text interpretiert (bis zum RETURN)
			while((!(rs232_buf[chr_out] == 13)) && (chr_out <= chr_in))
				{
				if (rs232_buf[chr_out] == 92) // wenn ein Backslash im Puffer steht
					{
			      if (chr0_flag)          // wenn das Flag bereits gesetzt, dann einen CHR(0) schreiben
			      	{
			      	write_buf[chr_twi] = 0; // ein CHR(0) schreiben als Abschluss eines Strings
			      	chr_twi++;
			      	chr_out++;           // jetzt sollte eigentlich ein RETURN folgen - wird hier nicht geprft

			      	goto exit_read;	   // das weitere Einlesen wird abgebrochen, eine '0' innerhalb eines Textes macht keinen Sinn
			      	}

			      else							// wenn das Flag noch nicht gesetzt (am Anfang des Textblockes)
			      	{
			      	chr0_flag++;			// Flag setzen
			      	chr_out++;				// Zhler incrementieren, aber ohne eine Ausgabe in den Sendepuffer

			      								// damit ein einzelner Backslash am Ende einer Zeile ausreicht, um ein '0' zu senden
													// prfen, ob das Ende der Eingabe erreicht ist
			      	if (chr_out == (chr_in - 1))
			      		{
			      		write_buf[chr_twi++] = 0;
			      		goto exit_read;	// Einlesen beenden
			      		}
			     		}
					}

				else            				// wenn kein Backslash im Puffer, dann Zeichen einfach im Sendepuffer ablegen
					{
					write_buf[chr_twi++] = rs232_buf[chr_out++];
					}
		  		}
			}
		}

	exit_read:									// nicht 100%ig elegant ....

// Ausgabe auf den TWI- Bus ---------------------------------------------------
	if (chr_out)								// nur weiter, wenn > 0 (wenn berhaupt Eingaben vorhanden)
		{
		if (write_buf[0] & 0x01)			// testen, ob Adressbit.0 gesetzt, dann wird vom TWI-Bus gelesen
			{
			if (chr_twi == 2)					// wenn genau 2 Byte eingelesen, dann sind das die Adresse + Anzahl der zu lesenden Bytes
				{
				chr_twi = write_buf[1] + 1;// fr die Adresse um 1 erhhen
													// aber die Arraygrenzen beachten !!
				if (chr_twi > MAX_BUFFER) chr_twi = MAX_BUFFER;
				}

			else if (chr_twi < 2)   	// sonst wird die Anzahl der Parameter gewhlt
				{
				chr_twi = 2;
				}

			else
				{
				chr_twi ++;						// die Anzahl der Bytes + 1
				}
													// um die Eingabe nicht zu berschreiben, wird zum Empfang ein anderer Buffer
													// verwendet. Dadurch kann mit +++ die letzte Eingabe bernommen werden.
			read_buf[0] = write_buf[0];	// write_buf nach read_buf kopieren

			TWI_Start_Transceiver_With_Data(read_buf, chr_twi);

													// wenn Daten empfangen wurden
			if (TWI_Get_Data_From_Transceiver(read_buf, chr_twi))
				{
				puts_p(PSTR("Rxd "));
				put_string(itoa(read_buf[0], msg, format));
				puts_p(PSTR(": "));
				for(i = 1; i< chr_twi; i++)
					{
					put_string(itoa(read_buf[i], msg, format));
					put_c(',');
					}
					put_cr();
				}

			else  									// wenn keine Daten empfangen wurden
				{
				temp = TWI_Get_State_Info();
				puts_p(PSTR("Adr "));			// Gerteadresse ausgeben
				put_string(itoa(read_buf[0], msg, format));
				err_msg(temp);						// Fehlercode/meldung ausgeben incl. Return
				}
			}

		else											// Adressbit.0 war 0, es wird geschrieben
			{
			TWI_Start_Transceiver_With_Data(write_buf, chr_twi);

			if (no_echo == 0)   					// dann ein Echo ausgeben
				{
				temp = TWI_Get_State_Info();	// temp zum Zwischenspeichern des Fehlercodes
				if (temp == TWI_NO_STATE)		// kein Fehler
					{
					puts_p(PSTR("Txd "));		// Gerteadresse ausgeben
					put_string(itoa(write_buf[0], msg, format));
					puts_p(PSTR(": "));
					for(i = 1; i < chr_twi; i++)
						{
						put_string(itoa(write_buf[i], msg, format));
						put_c(',');
						}
					put_cr();
					}

				else									// Fehlercode und Gerteadresse ausgeben
					{
					puts_p(PSTR("Adr "));
					put_string(itoa(write_buf[0], msg, format));
					err_msg(temp);					// Fehlercode/meldung ausgeben incl. Return
					}
				}
			}
		}
	}
}

// ----------------------------------------------------------------------------
// gibt ein einzelnes Zeichen auf COM aus
void put_c(uint8_t Zeichen)
{
while(!(UCSRA &(1<<UDRE)));						// warten bis sendebereit
UDR = Zeichen;
}

// ----------------------------------------------------------------------------
// gibt ein einzelnes Zeichen auf COM aus
void put_cr(void)
{
while(!(UCSRA &(1<<UDRE)));						// warten bis sendebereit
UDR = 13;
}

// ----------------------------------------------------------------------------
// gibt einen String aus dem RAM aus
void put_string(char *s)
{
while(*s)
	{
	put_c(*s);
	s++;
	}
}

// ----------------------------------------------------------------------------
// gibt einen String aus dem Flash aus
void puts_p(const char *progmem_s)
{
register char c;
while ( (c = pgm_read_byte(progmem_s++)) )
	{
   put_c(c);
	}
}

// ----------------------------------------------------------------------------
// zeigt den status von no_echo aus
// (Echo der Eingabe ber COM zurcksenden ans Terminal)
void show_echo(void)
{
puts_p(PSTR("Tx-Echo: "));					// und Status ausgeben
if (no_echo)
	{
	puts_p(PSTR("AUS (fr Datei-Input)"));
	}
else
	{
	puts_p(PSTR("EIN"));
	}
put_cr();
}

// ----------------------------------------------------------------------------
// gibt das Ausgabeformat aus, 'format' ist wird als Parameter in 'itoa()' benutzt
void show_format(void)
{
puts_p(PSTR("Ausgabeformat: "));
if (format == 2)
	{
	puts_p(PSTR("binaer"));
	}
else if (format == 10)
	{
	puts_p(PSTR("dezimal"));
	}
else if (format == 16)
	{
	puts_p(PSTR("hexadezimal"));
	}
put_cr();
}

// ----------------------------------------------------------------------------
// gibt den TWI-Takt aus
void show_twi_takt(uint8_t takt)
{
puts_p(PSTR("TWI-Takt = "));
put_string(itoa(takt, msg, 10));
puts_p(PSTR("KHz, TWBR = "));
put_string(itoa(TWBR, msg, 10));
put_cr();
}


// ----------------------------------------------------------------------------
// Meldung, die bei Einschalten ausgeben wird (und bei -?)
void help(void)
{
put_cr();
put_cr();
puts_p(PSTR("TERMINAL_2_TWI"));
put_cr();
put_cr();
puts_p(PSTR("Eingabe:"));
put_cr();
puts_p(PSTR("255(dez.Byte), 0xFF(hex.Byte), 0b11111111(bin.Byte), 0w65536(dez.Word)"));
put_cr();
put_cr();
puts_p(PSTR("Befehle:"));
put_cr();
puts_p(PSTR("-[0|1|2|3|4|5] -> TWBR einstellen"));
put_cr();
puts_p(PSTR("-x      -> TWI-Takt auf 100KHz einstellen"));
put_cr();
puts_p(PSTR("-[b|d|h]-> Ausgabe [binaer|dezimal|hexdezimal] -> "));
show_format();
puts_p(PSTR("-l      -> Anzeige der gefundenen TWI-Devices"));
put_cr();
#ifdef TSL2550
puts_p(PSTR("-t      -> Abfrage Lichtsensor TSL2550"));
put_cr();
#endif
puts_p(PSTR("-e      -> toggle "));				// den aktuellen Status schreibt show_echo()
show_echo();
puts_p(PSTR("           Fuer die Eingabe aus Text-Dateien AUS schalten !!"));
put_cr();
puts_p(PSTR("-?      -> diese Hilfe"));
put_cr();
put_cr();
}

// ----------------------------------------------------------------------------
// gibt Fehlercodes und Fehlermeldungen des TWI-Master aus
void err_msg(uint8_t error)
{
puts_p(PSTR(" TWI-Error 0x"));				//
put_string(itoa(error, msg, 16));			// Fehlercode immer hexadezimal
puts_p(PSTR(": "));								// Trennzeichen

if (error ==  0x08)
	{
	puts_p(PSTR("START has been transmitted"));
	}
else if (error ==  0x10)
	{
	puts_p(PSTR("Repeated START has been transmitted"));
	}
else if (error == 0x38)
	{
	puts_p(PSTR("Arbitration lost"));
	}
else if (error == 0x18 )
	{
	puts_p(PSTR("SLA+W has been transmitted and ACK received"));
	}
else if (error == 0x20 )
	{
	puts_p(PSTR("SLA+W has been transmitted and NACK received"));
	}
else if (error == 0x28 )
	{
	puts_p(PSTR("Data byte has been transmitted and ACK received"));
	}
else if (error ==  0x30)
	{
	puts_p(PSTR("Data byte has been transmitted and NACK received"));
	}
else if (error ==  0x40 )
	{
	puts_p(PSTR("SLA+R has been transmitted and ACK received"));
	}
else if (error == 0x48 )
	{
	puts_p(PSTR("SLA+R has been transmitted and NACK received"));
	}
else if (error == 0x50 )
	{
	puts_p(PSTR("Data byte has been received and ACK transmitted"));
	}
else if (error ==  0x58)
	{
	puts_p(PSTR("Data byte has been received and NACK transmitted"));
	}
//	else if (error ==  0xF8)	// ist das Ergebnis eines erfolgreichen Transfers
//		{
//		puts_p(PSTR("No relevant state information available; TWINT = 0"));
//		}
else if (error ==  0x00)
	{
	puts_p(PSTR("Bus error due to an illegal START or STOP condition"));
	}
else
	{
	puts_p(PSTR("unbekannter Fehler"));
	}
put_cr();
}
// ----------------------------------------------------------------------------
// Zeigt eine Liste aller am TWI-Bus antwortenden Gerte an
// wird durch die Eingabe von '-l' aufgerufen
void twi_test(void)
{
uint8_t i = 0;
uint8_t found = 0;

put_cr();
puts_p(PSTR("Suche nach TWI-Device ..."));
put_cr();

for (i=0; i < 128; i++)									// Adressen von 0 ... 254 Step 2
	{
	write_buf[0] = i<<1;									// Adresse in den Sendepuffer schreiben
//	write_buf[0] = write_buf[0] << 1;				// um eine Stelle nach links schieben

	TWI_Start_Transceiver_With_Data(write_buf, 1 );	// ein Byte (Adresse) senden
	_delay_ms(1);

																// Ergebnis testen auf TWI_NO_STATE
	if (TWI_Get_State_Info() == TWI_NO_STATE)		// dann liegt kein Fehler vor
		{
		found++;												// Zhler fr gefundene Gerte
		if (format == 2) puts_p(PSTR("0b"));
		if (format == 16) puts_p(PSTR("0x"));
		put_string(itoa(write_buf[0], msg, format));// Adresse ausgeben
		put_c(',');											// Komma + Leerzeichen ausgeben
		put_c(' ');
		}
	}

if (found)													// Zusammenfassung am Ende
	{
	put_cr();												// neue Zeile
	puts_p(PSTR("gefunden: "));
	put_string(itoa(found, msg, 10));				// Print Anzahle Gerte gefunden
	}

else
	{
	put_cr();												// neue Zeile
	puts_p(PSTR("kein TWI-Device gefunden!"));	// Print keine Gerte gefunden
	}

put_cr();													// neue Zeile
}

// ----------------------------------------------------------------------------
// liest eine Ziffernfolge aus dem Puffer und gibt sie als ein Byte zurck
// die aufrufende Routine hat auf isdigit() geprft.
// die Ziffernfolge kann mit '0b' oder '0x' beginnen, dann wird in Unterprogramme
// zur Erkennung verzweigt.

uint8_t Zahl_read(void)
{
uint8_t Zahl = 0;
uint8_t i = 0;
uint8_t zero = 0;											// wird TRUE, wenn die erste Ziffer eine 0 ist
																// fr die Eingabe im Format 0x../ 0b..
if (rs232_buf[chr_out] == '0') zero = TRUE;
do
	{
	if (isdigit(rs232_buf[chr_out]))					// solange einlesen, wie Ziffern vorhanden
		{
		Zahl *= 10;
		Zahl += (rs232_buf[chr_out++] - 0x30);
		i++;
		}
	else														// taucht ein Buchstabe auf, dann ...
		{
		if ((zero) && (i == 1))							// wenn das erste Zeichen eine '0' war und nun
			{													// das nchste (das zweite) Zeichen ansteht
			rs232_buf[chr_out] = toupper(rs232_buf[chr_out]); // in Groschreibung umwandeln
			if (rs232_buf[chr_out] == 'X')
				{
				return(hex_read());						// eine Hexadezimalzahl einlesen und zurckgeben
				}

			else if (rs232_buf[chr_out] == 'B')
				{
				return(bin_read());						// eine Binrzahl einlesen und zurckgeben
				}
																// Word bei 0l.. oder 0w...
			else if ((rs232_buf[chr_out] == 'W') || (rs232_buf[chr_out] == 'L'))
				{
				return(long_read());
				}
			}

		else													// ein Buchstabe taucht in der Ziffernfolge auf
			{
			break;											// abbrechen
			}
		}
	}
	while ((!(rs232_buf[chr_out] == ',')) && (!(rs232_buf[chr_out] == 13)) && (i < 3));
	// solange einlesen bis:
	// ',' oder RETURN eingelesen oder
	// mehr als drei Zeichen eingelesen (255!), berlauf wird ignoriert (256 wird zu 1)!!

	return(Zahl);
}

// ----------------------------------------------------------------------------
// liest eine Zahl vom Typ long/word mit bis zu 5 Stellen ein
// das high-byte schreibt die Funktion in den TWI-Buffer, das low-Byte wird zurckgegeben
uint8_t long_read(void)
{
uint16_t Zahl = 0;
uint8_t i = 0;

chr_out++;													// von 0l.. zur ersten Ziffer gehen
do
	{
	if (isdigit(rs232_buf[chr_out]))					// solange einlesen, wie Ziffern vorhanden
		{
		Zahl *= 10;
		Zahl += (rs232_buf[chr_out++] - 0x30);
		i++;
		}
	else
		{
		break;
		}
	}
while ((!(rs232_buf[chr_out] == ',')) && (!(rs232_buf[chr_out] == 13)) && (i < 5));

write_buf[chr_twi] = Zahl >> 8;						// schreibt hier das high-Byte,
chr_twi++;

return(Zahl);												// das low-Byte wird an die aufrufende Funktion zurckgegeben
}

// ----------------------------------------------------------------------------
// liest die nachfolgenden (bis zu) 8 Stellen als Binrzahl ein
// ein Zeichen <>[0|1] beendet die Funktion
// der Index chr_out steht noch auf "B" von 0b...
uint8_t bin_read(void)
{
uint8_t bin = 0;
uint8_t i = 0;

for (i = 0; i < 8; i++)
	{
	chr_out++;												// Index incrementieren
	if (rs232_buf[chr_out] == '0')					// bei einer Null nur nach links schieben
		{
		bin *= 2;											// um eine Stelle nach links schieben
		}

	else if (rs232_buf[chr_out] == '1')				// bei einer Eins nach links schieben
		{														// und 1 addieren
		bin *= 2;											// um eine Stelle nach links schieben
		bin++;												// und eine 1 addieren
		}

	else														// wenn weder 0 noch 1, dann
		{
		return(bin);										// den bisher ermittelten Wert zurckliefern,
		}														// den Index chr_out nicht verndern
	}

chr_out++;													// den Index auf das nchste Zeichen setzen
return(bin);
}

// ----------------------------------------------------------------------------
// liest zwei folgende Zeichen als Hexadezimalzahl ein
// ein Zeichen, das weder eine Ziffer noch [A-E] ist, beendet die Eingaben
// der Index chr_out steht noch auf dem "X" von 0x..
uint8_t hex_read(void)
{
uint8_t hex = 0;
uint8_t i = 0;
int8_t temp;

for (i = 0; i < 2; i++)
	{
	chr_out++;												// Index von rs232_buf auf nchstes Zeichen
	temp = toupper(rs232_buf[chr_out]);			// in Groschreibung umwandeln und puffern
																// rs232_buf knnte ja einen nachfolgenden Text enthalten ...
//	if ((temp > 47) && (temp < 58))					// auf Ziffer prfen
	if (isdigit(temp))
		{
		hex *= 16;
		hex += (temp - 0x30);							// Wenn Ziffer, dann zu Zahl machen
		}

	else if ((temp > 64) && (temp < 71)) 			// Wenn 'A .. F'
		{
		hex *= 16;
		hex += (temp - 65 + 10);						// in Ziffer umwandeln und addieren
		}

	else
		{
		return(hex);										// den bisher ermittelten Wert zurckliefern,
		}														// den Index chr_out nicht verndern
	}

chr_out++;													// den Index auf das nchste Zeichen setzen
return(hex);												// Wert zurckliefern
}

